from functools import lru_cache
from typing import Iterable, Tuple

from onto.attrs import AttributeBase


def _make_schema_name(cls):
    return f"_{cls.__name__}_GeneratedSchema"


def _schema_cls_from_attributed_class(cls):
    """ Make schema from a class containing AttributeBase+ objects

    :return:
    """

    import_only = getattr(cls.Meta, "import_only", False)
    export_only = getattr(cls.Meta, "export_only", False)

    d = dict()
    for key, attr in _collect_attrs(cls):
        field = attr._make_field()
        field.load_only = field.load_only or import_only
        field.dump_only = field.dump_only or export_only
        d[key] = field
    if len(d) == 0:
        return None
    else:
        m_meta = {
            "exclude": getattr(cls.Meta, "exclude", tuple())
        }

        d["Meta"] = type(
            "Meta",
            tuple(),
            m_meta
        )

        if hasattr(cls.Meta, "case_conversion"):
            d["case_conversion"] = cls.Meta.case_conversion

        schema_base = cls._schema_base if cls._schema_cls is None else cls._schema_cls

        TempSchema = type(_make_schema_name(cls), (schema_base,), d)

        return TempSchema


def _graphql_type_from_py(t: type, input=False):
    import graphql
    PY_TYPE_MAP_GQL = {
        int: graphql.GraphQLInt,
        bool: graphql.GraphQLBoolean,
        float: graphql.GraphQLFloat,
        str: graphql.GraphQLString,
        list: graphql.GraphQLList
    }
    if t in PY_TYPE_MAP_GQL:
        return PY_TYPE_MAP_GQL[t]
    else:
        if input:
            return _graphql_object_type_from_attributed_class(t, input=input)
        else:
            raise ValueError


def _graphql_field_from_attr(attr, input=False):
    import graphql

    if not input:
        field_base = graphql.GraphQLField
    else:
        field_base = graphql.GraphQLInputField

    if hasattr(attr, 'collection') and attr.collection == list:
        f = lambda ot: graphql.GraphQLList(ot)
    else:
        f = lambda ot: ot

    from onto import attrs
    if attr.__class__ is attrs.attribute.EmbeddedAttribute:
        e_cls = attr.type_cls
        e_graphql = _graphql_object_type_from_attributed_class(e_cls)
        field = field_base(
            type_=f(e_graphql),
            description=attr.doc
        )
        return attr.data_key, field
    elif isinstance(attr, attrs.attribute.AttributeBase):
        import graphql
        field = field_base(
            type_=f(_graphql_type_from_py(t=attr.type_cls)),
            description=attr.doc
        )
        return attr.data_key, field
    else:
        raise NotImplementedError


@lru_cache(maxsize=None)  # Cached property
def _graphql_object_type_from_attributed_class(cls, input=False):
    """ Make GraphQL schema from a class containing AttributeBase+ objects

    :return:
    """

    # import_only = getattr(cls.Meta, "import_only", False)
    # export_only = getattr(cls.Meta, "export_only", False)

    import graphql

    fields = dict(
        _graphql_field_from_attr(attr, input=input)
        for key, attr in _collect_attrs(cls) if not attr.is_internal
    )

    if not input:
        base = graphql.GraphQLObjectType
    else:
        base = graphql.GraphQLInputObjectType

    type_name = cls.__name__

    # TODO: maybe add
    # if input:
    #     type_name += "Input"

    graphql_object_type = base(
        type_name,
        fields=fields
    )

    return graphql_object_type


def _collect_attrs(cls) -> Iterable[Tuple[str, AttributeBase]]:
    """
    Collect all AttributeBase+ objects in the class and its ancestors.
    TODO: debug empty iter of length 0

    :param cls:
    :return:
    """
    for key in dir(cls):
        if issubclass(getattr(cls, key).__class__, AttributeBase):
            attr = getattr(cls, key)
            yield (key, attr)



from app import app

from onto.context import Context as CTX
application = CTX.services.engine.wrap(app)

if __name__ == '__main__':
    import os
    PORT = int(os.environ['LEANCLOUD_APP_PORT'])

    # 只在本地开发环境执行的代码
    from gevent.pywsgi import WSGIServer
    from werkzeug.serving import run_with_reloader
    from werkzeug.debug import DebuggedApplication

    @run_with_reloader
    def run():
        global application
        app.debug = True
        application = DebuggedApplication(application, evalex=True)
        server = WSGIServer(('0.0.0.0', PORT), application)
        server.serve_forever()

    run()




from typing import Any

from marshmallow.utils import get_value

from .attribute import PropertyAttribute, _ATTRIBUTE_STORE_NAME
from ..common import _NA


class Accumulator:

    @classmethod
    def create(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def __init__(self, value=None):
        self.value = value


class Aggregation(PropertyAttribute):
    """
    Used to aggregate multiple rows in a table to a single row
    """

    def __create_accumulator__(self, instance, value):
        inner = getattr(instance, _ATTRIBUTE_STORE_NAME)
        setattr(inner, self.name, self.f_create_accumulator(instance, value))

    def __init__(self, import_default=_NA, **kwargs):
        """ Never initialize; always set

        :param kwargs:
        """

        """
        Getter: returns accumulator
        """
        def fget(_self) -> Accumulator:
            inner = getattr(_self, _ATTRIBUTE_STORE_NAME)
            return getattr(inner, self.name)

        """
        Setter: creates accumulator
        """
        def fset(_self, value):
            self.__create_accumulator__(_self, value)

        """
        Deleter: deletes accumulator
        """
        def fdel(_self):
            inner = getattr(_self, _ATTRIBUTE_STORE_NAME)
            return delattr(inner, self.name)

        super().__init__(
            initialize=False,
            import_enabled=True,
            import_default=import_default,
            fget=fget,
            fset=fset,
            fdel=fdel,
            **kwargs
        )

        def create_accumulator(_self, value=_NA):
            return Accumulator.create(value=value)

        self.f_create_accumulator = create_accumulator

        def f_accumulate(_self):
            raise NotImplementedError

        self.f_accumulate = f_accumulate

    def create_accumulator(self, f_create_accumulator):
        self.f_create_accumulator = f_create_accumulator

    def accumulate(self, f_accumulate):
        self.f_accumulate = f_accumulate




from onto.models.utils import _schema_cls_from_attributed_class
from onto.registry import ModelRegistry


# DEFAULT_FIELDS = {"obj_type", "doc_id", "doc_ref"}


class SerializableMeta(ModelRegistry):
    """
    Metaclass for serializable models.
    """

    def __new__(mcs, name, bases, attrs):
        klass = super().__new__(mcs, name, bases, attrs)

        meta = klass.Meta
        if hasattr(meta, "schema_cls"):
            klass._schema_cls = meta.schema_cls

        attributed = _schema_cls_from_attributed_class(cls=klass)
        if attributed is not None:
            klass._schema_cls = attributed
        # if hasattr(klass, "Meta"):
        #     Moves Model.Meta.schema_cls to Model._schema_cls


            # if hasattr(meta, "default_fields"):
            #     default_fields = meta.default_fields
            #     if default_fields > DEFAULT_FIELDS:
            #         raise ModelDeclarationError(
            #             f"default_fields argument: {default_fields} "
            #             f"in {name}.Meta is not a subset (<=) of "
            #             f"{DEFAULT_FIELDS} "
            #         )
            #     klass._default_fields = default_fields

        return klass




import pytest

from tests.utils import _delete_all

def pytest_sessionstart(session):
    from onto.context import Context as CTX
    from onto.config import Config

    if CTX.config is None:
        config = Config.load()
        CTX.read(config)

    _delete_all(CTX, "Orbit")
    _delete_all(CTX, "locations")
    _delete_all(CTX, "orbits")
    _delete_all(CTX, "rideHosts")
    _delete_all(CTX, "riderBookings")
    _delete_all(CTX, "riderTargets")
    _delete_all(CTX, "users")

    _delete_all(CTX, "RainbowDAV")
    _delete_all(CTX, "Shard")
    _delete_all(CTX, "UserViewDAV")
    _delete_all(CTX, "categories")
    _delete_all(CTX, "counters")
    _delete_all(CTX, "hellos")

    _delete_all(CTX, subcollection_name="MeetingSessionDAV_PATCH")
    _delete_all(CTX, subcollection_name="_PATCH_UserViewDAV")
    _delete_all(CTX, subcollection_name="MeetingSessionDAV")


def pytest_sessionfinish(session, exitstatus):
    from onto.context import Context as CTX

    if exitstatus == 0:
        _delete_all(CTX, "Orbit")
        _delete_all(CTX, "locations")
        _delete_all(CTX, "orbits")
        _delete_all(CTX, "rideHosts")
        _delete_all(CTX, "riderBookings")
        _delete_all(CTX, "riderTargets")
        _delete_all(CTX, "users")

        _delete_all(CTX, "RainbowDAV")
        _delete_all(CTX, "Shard")
        _delete_all(CTX, "UserViewDAV")
        _delete_all(CTX, "categories")
        _delete_all(CTX, "counters")
        _delete_all(CTX, "hellos")

        _delete_all(CTX, subcollection_name="MeetingSessionDAV_PATCH")
        _delete_all(CTX, subcollection_name="_PATCH_UserViewDAV")
        _delete_all(CTX, subcollection_name="MeetingSessionDAV")



from onto.context import Context as CTX

CTX.load()

engine = CTX.services.engine


from onto.view import Mediator

# JSONRPC_URI = 'http://jsonrpc.multiri.de'
JSONRPC_URI = 'http://0.0.0.0:5000'

class TodoMediatorLc(Mediator):

    from onto.source.leancloud import hook
    from onto.sink.json_rpc import sink

    src = hook('Todo')
    snk = sink(uri=f'{JSONRPC_URI}/todo')

    @src.triggers.after_save
    def call_after_save_rpc(self, ref, snapshot):
        self.snk.emit('after_save', ref=str(ref), snapshot=snapshot.to_dict())
    #
    # @src.triggers.before_save
    # def fb_before_todo_save(self, ref, snapshot):
    #     from onto.database.firestore import FirestoreReference
    #     CTX.dbs.firestore.set(ref=FirestoreReference.from_str(str(ref)), snapshot=snapshot)
    #     raise ValueError(f"{str(ref)} {str(snapshot)}")

    @classmethod
    def start(cls):
        cls.src.start()

#
# class TodoMediator(Mediator):
#
#     from onto.source.json_rpc import JsonRpcSource as source
#     src = source(url_prefix='todo')
#
#     @src.triggers.after_save
#     def record_todo(self, response):
#         from jsonrpcclient import Response
#         assert isinstance(response, Response)
#         if response.data.ok:
#             res = response.data.result
#             ref = res['ref']
#             snapshot = res['snapshot']
#             self._fb_before_todo_save(ref=ref, snapshot=snapshot)
#         else:
#             raise Exception(response.raw)
#
#     def _fb_before_todo_save(self, ref, snapshot):
#         from onto.database.firestore import FirestoreReference
#         CTX.dbs.firestore.set(ref=FirestoreReference.from_str(str(ref)), snapshot=snapshot)
#         raise ValueError(f"{str(ref)} {str(snapshot)}")



from onto import domain_model, attrs
# from . import User

class TicketBase(domain_model.DomainModel):

    class Meta:
        collection_name = "tickets"


class Ticket(TicketBase):
    role = attrs.bproperty()
    user = attrs.relation(nested=False, dm_cls='User')
    attendance = attrs.bproperty()




from .meta import SerializableMeta
from ..registry import ModelRegistry
from .mixin import Importable, NewMixin, Exportable
from .utils import _collect_attrs
from onto.mapper.schema import Schema


class SimpleStore:
    """
    To store simple business properties
    """
    pass


class BaseRegisteredModel(metaclass=ModelRegistry):
    """
    Ref: https://github.com/faif/python-patterns/blob/master/patterns/behavioral/registry__py3.py
    """

    @classmethod
    def _get_children(cls):
        return {cls.get_cls_from_name(c_str)
                for c_str in cls._get_children_str(cls.__name__)}

    @classmethod
    def _get_subclasses(cls):
        res = {cls, }
        children = cls._get_children()
        for child in children:
            res |= child._get_subclasses()
        return res

    @classmethod
    def _get_subclasses_str(cls):
        return list(
            sorted(_cls.__name__ for _cls in cls._get_subclasses())
        )

    @classmethod
    def _get_parents(cls):
        return {cls.get_cls_from_name(c_str)
                for c_str in cls._get_parents_str(cls.__name__)}


class SchemedBase:

    @property
    def schema_obj(self):
        raise NotImplementedError

    @property
    def schema_cls(self):
        raise NotImplementedError


class Schemed(SchemedBase):
    """
    A mixin class for object bounded to a schema for serialization
        and deserialization.
    """

    _schema_obj = None
    _schema_cls = None

    # def __init__(self, *args, **kwargs):
    #     super().__init__(*args, **kwargs)
    # self._schema_obj = self._schema_cls()

    @classmethod
    def get_schema_cls(cls):
        """ Returns the Schema class associated with the model class.
        """
        return cls._schema_cls

    @classmethod
    def get_schema_obj(cls):
        """ Returns an instantiated object for Schema associated
                with the model class
        """
        # Use __dict__ to avoid reading from super class
        if "_schema_obj" not in cls.__dict__:
            schema_cls = cls.get_schema_cls()
            if schema_cls is None:
                return None
            cls._schema_obj = schema_cls()
        return cls._schema_obj

    @property
    def schema_cls(self):
        """ Returns the Schema class associated with the model object.
        """
        return self.get_schema_cls()

    @property
    def schema_obj(self):
        """ Returns an instantiated object for Schema associated
                with the model object.
        """
        return self.get_schema_obj()

    @classmethod
    def _get_fields(cls):
        fd = cls.get_schema_obj().fields
        return {
            key: val for key, val in fd.items()  # TODO: change
        }

    #     """ TODO: find ways of collecting fields without reading
    #                 private attribute on Marshmallow.Schema
    #
    #     :return:
    #     """
    #     res = dict()
    #     for name, declared_field in cls.get_schema_obj().fields.items():
    #         if not declared_field.dump_only:
    #             res[name] = declared_field
    #     return res


class Mutable(BaseRegisteredModel,
              Schemed, Importable, NewMixin, Exportable):
    pass


class Immutable(BaseRegisteredModel, Schemed, NewMixin, Exportable):
    pass


class Serializable(Mutable, metaclass=SerializableMeta):

    class Meta:
        pass
    #     """
    #     Options object for a Serializable model.
    #     """
    #     schema_cls = None

    _schema_base = Schema

    def _init__attrs(self):
        self._attrs = SimpleStore()
        for key, attr in _collect_attrs(cls=self.__class__):
            if attr.initialize:
                attr.initializer(self)

    def __init__(self, *args, **kwargs):
        self._init__attrs()
        super().__init__(*args, **kwargs)



from onto import domain_model, attrs


class LocationBase(domain_model.DomainModel):

    class Meta:
        collection_name = "locations"

    latitude = attrs.bproperty()
    longitude = attrs.bproperty()
    address = attrs.bproperty()


class Location(LocationBase):

    pass



from onto import domain_model, attrs
# from . import Location, User, Ticket


class MeetingBase(domain_model.DomainModel):

    class Meta:
        collection_name = "meetings"


class Meeting(MeetingBase):

    location = attrs.relation(nested=False, dm_cls='Location')
    users = attrs.relation(nested=False, dm_cls='User',
                           collection=list)
    tickets = attrs.relation(nested=False, dm_cls='Ticket',
                             collection=dict)
    status = attrs.bproperty()



from collections import namedtuple

from .attribute import AttributeBase, \
    PropertyAttribute as bproperty, \
    RelationshipAttribute as relation, \
    DictAttribute as bdict, \
    EmbeddedAttribute as embed, \
    ObjectTypeAttribute as object_type, \
    LocalTimeAttribute as local_time, \
    PropertyAttribute as string, \
    PropertyAttribute as integer, \
    DocRefAttribute as doc_ref, \
    PropertyAttribute as action

# TODO: change import for string and integer



from examples.meeting_room.views.graphql_view import UserGraphQLMediator, UserLoginMediator
from onto.sink.graphql import op_schema

schema_all = list()

schema_all += UserGraphQLMediator.start()
schema_all += UserLoginMediator.start()
#
from graphql import GraphQLSchema

schema = GraphQLSchema(
    query=op_schema(op_type='Query', schema_all=schema_all),
    subscription=op_schema(op_type='Subscription', schema_all=schema_all),
    mutation=op_schema(op_type='Mutation', schema_all=schema_all)

)

from stargql import GraphQL

app = GraphQL(
    schema=schema,
)

# PYTEST FIXTURES: RETAIN
from tests.fixtures import CTX
from examples.meeting_room.tests.fixtures import users, meeting, location, tickets


def test_run(users, meeting, location, tickets):

    import asyncio
    loop = asyncio.get_event_loop()

    from uvicorn import Config
    config = Config(app=app, loop=loop, port=8080, debug=True)
    from uvicorn import Server
    server = Server(config)
    loop.run_until_complete(server.serve())

    # uvicorn.run(app, port=8080, debug=True)



# import typing
#
# import pytest
# from marshmallow import schema, fields, post_load
#
#
# def test_stuffs():
#
#     class ReferenceField(fields.Field):
#
#         def _deserialize(
#         self,
#         value: typing.Any,
#         attr: typing.Optional[str],
#         data: typing.Optional[typing.Mapping[str, typing.Any]],
#         **kwargs
#     ):
#             dm_cls = self.context['dm_cls']
#             general_schema = dm_cls.schema_cls()
#
#             obj_type = general_schema.find_obj_type(value)
#
#             class TmpSchema(schema.Schema):
#                 d = fields.Nested(obj_type.schema_cls)
#
#             this_schema = TmpSchema()
#             this_schema.context['obj_type'] = obj_type
#
#             this_schema.context['stuffs'] = self.context['stuffs']
#             return this_schema.load(dict(d=value))
#
#     class FirestoreSnapshotSchema(schema.Schema):
#         d = ReferenceField()
#
#     class Location:
#
#         @classmethod
#         def from_dict(cls, d):
#             src = dict(d=d)
#             master_schema = FirestoreSnapshotSchema()
#             master_schema.context['dm_cls'] = cls
#             return master_schema.load(data=src)
#
#
#     class LocationSchema(schema.Schema):
#         sublocation = ReferenceField()
#
#         @post_load
#         def make_object(self, data, **kwargs):
#             obj_type = self.context['obj_type']
#             return obj_type.new(**data)
#



from functools import partial
from pathlib import Path
from subprocess import Popen

from pytest_kafka import (
    make_zookeeper_process, make_kafka_server, make_kafka_consumer,
    terminate,
)

ROOT = Path.cwd()
KAFKA_SCRIPTS = ROOT / 'pytest-kafka-0.4.0' / 'kafka/bin/'
KAFKA_BIN = str(KAFKA_SCRIPTS / 'kafka-server-start.sh')
ZOOKEEPER_BIN = str(KAFKA_SCRIPTS / 'zookeeper-server-start.sh')
# You can pass a custom teardown function (or parametrise ours). Just don't call it `teardown`
# or Pytest will interpret it as a module-scoped teardown function.
teardown_fn = partial(terminate, signal_fn=Popen.kill)
zookeeper_proc = make_zookeeper_process(ZOOKEEPER_BIN, teardown_fn=teardown_fn)
kafka_server = make_kafka_server(KAFKA_BIN, 'zookeeper_proc', kafka_port=9092, teardown_fn=teardown_fn)
kafka_consumer = make_kafka_consumer(
    'kafka_server', seek_to_beginning=True, kafka_topics=['topic'])


def test_fixture(kafka_server):
    from kafka import KafkaProducer

    # assert isinstance(kafka_server, KafkaProducer)
    kafka_server = KafkaProducer(bootstrap_servers='localhost:9092', api_version=(0, 10, 1))
    kafka_server.send('my-topic', value='hello world!'.encode('utf-8'))



""" DocumentMediator listens to change in a single document and triggers
        functions when such document change.
"""
from .document import ViewMediatorDAV as DocumentMediator

""" QueryMediator listens to the result of a query and triggers 
        functions when the result of such query changes. 
"""
from .query_delta import ViewMediatorDeltaDAV as QueryMediator, \
    ProtocolBase, \
    OnSnapshotTasksMixin

""" RestMediator pulls and modifies data when requested with REST API. 
"""
from .rest_api import ViewMediator as RestMediator

""" WsMediator establishes websocket connection with client. 
"""
from .websocket import ViewMediatorWebsocket as WsMediator


class Mediator:

    source = None
    subscribe_user_view = None

    @classmethod
    def start(cls):
        cls.source.start()



from onto import domain_model, attrs


class UserBase(domain_model.DomainModel):
    pass


class User(UserBase):

    class Meta:
        collection_name = "users"

    first_name = attrs.bproperty()
    last_name = attrs.bproperty()
    organization = attrs.bproperty()
    hearing_aid_requested = attrs.bproperty()
    display_name = attrs.bproperty(import_enabled=False)

    @display_name.getter
    def display_name(self):
        return "{} {}".format(self.first_name, self.last_name)



import uvicorn as uvicorn
from examples.meeting_room.domain_models import User

from examples.meeting_room.view_models import UserView
from onto import view_model, attrs
from onto.view import Mediator


class UserGraphQLMediator(Mediator):

    from onto.source import domain_model
    from onto.sink.graphql import subscription

    src = domain_model(domain_model_cls=User)
    subscribe_user_view = subscription(view_model_cls=UserView)

    @subscribe_user_view.triggers.add_topic
    def add_topic(self, user_id: str):
        return user_id

    @subscribe_user_view.triggers.on_event
    def on_event(self, event: dict):
        return event

    @src.triggers.on_create
    @src.triggers.on_update
    def user_updated(self, obj: User):
        v = UserView.get(user_id=obj.doc_id)
        d = v.to_dict()
        q = self.subscribe_user_view.qs[obj.doc_id]

        import asyncio
        # tell asyncio to enqueue the result
        fut = asyncio.run_coroutine_threadsafe(
            q.put(d), loop=self.subscribe_user_view.loop
        )
        # wait for the result to be enqueued
        _ = fut.result()

    @classmethod
    def start(cls):
        cls.src.start()
        s = cls.subscribe_user_view.start()
        from onto.sink.graphql import graph_schema
        import graphql
        liveness = graph_schema(
            op_type='Query',
            name='liveness',
            graphql_object_type=graphql.GraphQLObjectType(
                name='Liveness',
                fields={
                    'alive': graphql.GraphQLField(
                        graphql.GraphQLBoolean,
                        resolve=lambda *args, **kwargs: True),
                }
            ),
            args=dict()
        )

        return [s, liveness]



class LoginForm(view_model.ViewModel):
    phone_number = attrs.bproperty(type_cls=str)
    verification_code = attrs.bproperty(type_cls=str)


class LoginRes(view_model.ViewModel):
    token = attrs.bproperty(type_cls=str)


class UserLoginMediator(Mediator):

    from onto.sink.graphql import mutation

    login_user_view = mutation(view_model_cls=LoginRes)

    @login_user_view.triggers.mutate
    def login(self, form: LoginForm):
        print(form)

    @classmethod
    def start(cls):
        s = cls.login_user_view.start()
        return [s]


